require 'json'
require 'uu/os/vuc/component_container'
require 'uu/os/lang/uniform_hash'

module UU
  module OS
    module VUC

      # Class is used for building view response. Use ViewModel if you want get components, add message, change
      # view template etc.
      # @example
      #   def on_init(ctx)
      #
      #     # get viewModel class from context
      #     view = ctx.view
      #
      #     # get attribute
      #     format = view[:exportFormat] # String 'PDF'
      #
      #     # set attribute
      #     view[:exportFormat] = 'PDF'
      #
      #     # get component with code 'FORM_TEXT_AREA'
      #     view.component(:FORM_TEXT_AREA) # UU::OS::Form::Textarea
      #
      #     # get all components, whose code ends at 'TEXT_AREA'
      #     view.find_components(".*TEXT_AREA$") # Array [UU::OS::Form::Textarea, UU::OS::Form::Textarea]
      #
      #     info_msg = UU::OS::GVC::Message.new(
      #         {
      #             code: "CODE",
      #             severity: UU::OS::GVC::Message::MessageSeverity::INFO,
      #             text: "Info message."
      #         }
      #       )
      #     # add message to response
      #     view.add_message(info_msg)
      #
      #     # get all messages
      #     view.messages # Array [UU::OS::GVC::Message, UU::OS::GVC::Message]
      #
      #     # set focus on component with code 'FORM_TEXT_AREA'
      #     view.focus_component(:FORM_TEXT_AREA)
      #
      #     # you must return this object as result so all changes will be proceed
      #     return view
      #   end
      # @example
      #   def on_before_init(ctx)
      #     # get viewModel class from context
      #     view = ctx.view
      #
      #     # set template uri to 'ues:UNI:VUC:EXTENDED_FORM'
      #     view.template_uri("ues:UNI:VUC:EXTENDED_FORM")
      #
      #     return view
      #   end
      #
      # @example
      #   def on_value_change(ctx)
      #     # get viewModel class from context
      #     view = ctx.view
      #
      #     # get component which trigger event
      #     component = view.event_source_component # UU::OS::Form::Textarea
      #
      #   end
      class ViewModel

        # @private
        RESPONSE_JSON_SCHEMA_URI = "ues:SYSTEM:UU.OS.VUC/RESPONSE_MESSAGE-SCHEMA-V3" # :nodoc:

        # @param [UU::OS::VUC::VisualUseCaseContext] ctx Visual use case context.
        def initialize(ctx)
          @ctx = ctx
          @raw_attrs = UU::OS::Lang::UniformHash.new
        end

        # Returns attribute's value.
        #
        # @note Attribute's key have some restrictions:
        #   * Key of the attribute must not start with 'uu' prefix (e.g. uuName, uusage), prefix 'uu' is reserved for
        #     internal usage.
        #   * Key of the attribute must not be same as code of some component. It may cause conflict and the component
        #     will not work as expected. E.g. you have component with code 'NAME_INPUT' so you can not use this
        #     code as attribute code.
        #
        # @example
        #   # get viewModel class from context
        #   view = ctx.view
        #
        #   format = view[:exportFormat] # String 'PDF'
        #
        # @param [Symbol] key Attribute key in symbol.
        # @return [Object] Attribute value.
        def [](key)
          @raw_attrs[key]
        end

        # Set attribute's value. Attributes has a Use Case scope. It's means that attribute will be available in all
        # following actions on same Controller. E.g. If you set attribute in on_init(ctx) method, you can get this
        # attribute in on_value_change(ctx) or on_submit(ctx) actions. Attributes are not forwarded to next Uca Case
        # or between modal dialogs.
        #
        # @note Attribute's key have some restrictions:
        #   * Key of the attribute must not start with 'uu' prefix (e.g. uuName, uusage), prefix 'uu' is reserved for
        #     internal usage.
        #   * Key of the attribute must not be same as code of some component. It may cause conflict and the component
        #     will not work as expected. For example you have component with code 'NAME_INPUT' so you can not use this
        #     code as attribute code.
        #
        # @example
        #   # get viewModel class from context
        #   view = ctx.view
        #
        #   view[:exportFormat] = 'PDF'
        #
        #
        # @param [Symbol] key Attribute key in symbol.
        # @param [Object, nil] value Attribute value or nil if attribute with key do not exists.
        def []=(key, value)
          @raw_attrs[key] = value
        end

        # Returns component with code passed as parameter or nil if component with this code does not exists.
        #
        # @example
        #   # get viewModel class from context
        #   view = ctx.view
        #
        #   # get component with code 'FORM_TEXT_AREA'
        #   text_area = view.component(:FORM_TEXT_AREA) # UU::OS::Form::Textarea
        #
        #   # get non existing component
        #   view.component(:NON_EXISTS_CODE) # nil
        #
        # @param [Symbol] code Component code as symbol.
        # @return [UU::OS::GVC::Component, nil] Returns component by code or nil if component
        #    with this code does not exists.
        def component(code)
          code = code.to_sym if code.kind_of? String
          container.component_by_code(code)
        end

        # Search for component by regular expression or filtering block. If both expression and block are used, so block
        # has bigger priority and expression is discarded. If no expression or block are available so all components are
        # included in result. If expression or block match no component, empty array is returned.
        #
        # @note It is forbidden to modify components in filtering block.
        #
        # @example
        #   # get viewModel class from context
        #   view = ctx.view
        #
        #   # get all components which code starts with 'LABEL', e.g. 'LABEL_NAME', 'LABEL_SURNAME' but not 'NAME_LABEL'
        #   view.find_components("^LABEL") # Array [UU::OS::Form::Label, UU::OS::Form::Textarea]
        #
        #   # same as above but with Regexp object
        #   view.find_components(/^LABEL/)
        #
        #   # get all components
        #   view.find_components # Array [CompA, CompB, CompC]
        #
        #   # same as above
        #   view.find_components(nil) # Array [CompA, CompB, CompC]
        #
        #   # if expression match no component, empty array is returned
        #   view.find_components("NON_EXISTS") # Array []
        #
        #   # filter components using block, get all component with textarea type
        #   view.find_components do |comp|
        #     comp.component_type == 'uu.os.form.textarea'
        #   end
        #
        #   # get all read only textarea
        #   view.find_components do |comp|
        #     (comp.component_type == 'uu.os.form.textarea') && (comp.read_only)
        #   end
        #
        # @param [String, Regexp] code_regex Regular expression, all components which code match this expression will be
        #   returned in result.
        # @yield [comp] Block with one parameter that is actual component, if block returns false or nil so component
        # will not be included in result. Otherwise component will be part of result.
        # @yieldparam comp [UU::OS::GVC::Component] Actual component.
        # @return [Array] Returns array of filtered components.
        # @see #Regexp
        def find_components(code_regex = nil, &block)
          result = []

          if block_given?
            comps_list = container.components
            comps_list.each do |comp|
              include_comp = yield(comp)
              result << comp if include_comp
            end
          elsif !code_regex.nil?
            code_regex = code_regex.to_s
            comps_hash = container.get_components
            comps_hash.each do |key, comp|
              if key.to_s.match(code_regex)
                result << comp
              end
            end
          else
            result = container.components
          end

          return result
        end

        # Add user message, which will be displayed on form after this action.
        #
        # @example
        #
        #   info_msg = UU::OS::GVC::Message.new(
        #       {
        #           code: "CODE",
        #           severity: UU::OS::GVC::Message::MessageSeverity::INFO,
        #           text: "Info message."
        #       })
        #
        #   # add message as UU::OS::GVC::Message object
        #   view.add_message(info_msg)
        #
        #   msg_hash = {
        #       code: "CODE",
        #       severity: UU::OS::GVC::Message::MessageSeverity::INFO,
        #       text: "Info message."
        #   }
        #
        #   # add message as Hash
        #   view.add_message(msg_hash)
        #
        # @param [UU::OS::GVC::Message, Hash, String] message Message to be added. Message can be UU::OS::GVC::Message, Hash
        #   or JSON String.
        def add_message(message)
          container.add_message(message)
          return self
        end

        # Returns all messages added by {#add_message} method.
        #
        # @example
        #   # get all messages
        #   view.messages # Array [UU::OS::GVC::Message, UU::OS::GVC::Message]
        #
        # @return [Array] Array of UU::OS::GVC::Message.
        def messages
          container.messages.clone
        end

        # Change focus of components.
        #
        # @example
        #   # set focus on component with code 'FORM_TEXT_AREA'
        #   view.focus_component(:FORM_TEXT_AREA)
        #
        # @param [Symbol] code Component code, which will be focused.
        def focus_component(code)
          container.focus_by_code(code)
          return self
        end


        # Sets new view template. This template will be rendered as form. Default value is the main sheet of
        # registered VUC. You can use sheets from this VUC only, it's means that only sheets related to actual VUC artifact
        # are allowed, otherwise form execution fails with error. If you set non-exists uri than form execution will
        # fail as well.
        #
        # @note This parameters can be set only in on_before_init method, if you set template_uri in another
        # method, it will have no effect.
        #
        # @example
        #
        #   # uri of default VUC
        #   UCR_URI = "ues:UNI:CDS.GUESTBOOK/COMMENT/CREATE"
        #
        #   # code of sheet with extended form
        #   EXTENDED_FORM_CODE = "EXTENDED_FORM"
        #
        #   def on_before_init(ctx)
        #
        #     # get ViewModel instance from context
        #     view = ctx.view
        #
        #     use_extended_form = ... # some logic to find out if user can see extended form
        #
        #     if use_extended_form
        #       # set new sheet with extended form
        #       new_template = UU::OS::UESURIBuilder.parse_uesuri(UCR_URI).set_object_code(EXTENDED_FORM_CODE).to_uesuri
        #       view.template_uri(new_template)
        #     end
        #
        #     # if you do not set template_uri on view so default is used
        #
        #     return view
        #   end
        #
        # @param [UU::OS::UESURI, String] uu_uri UESURI of sheet that will be used as form template.
        def template_uri(uu_uri)
          @template_uri = uu_uri
          return self
        end

        # Returns component which trigger event. Result may be nil if event is not triggered by component. E.g. in
        # {UU::OS::VUC::VisualUseCaseController#on_init} method.
        #
        # @example
        #   def on_value_change(ctx)
        #     # get viewModel class from context
        #     view = ctx.view
        #
        #     # get component which trigger event
        #     component = view.event_source_component # UU::OS::Form::Textarea
        #     text = component.value
        #
        #   end
        #
        # @return [UU::OS::GVC::Component, nil] Returns component which trigger event or nil.
        def event_source_component
          uu_view = @ctx.parameters[:uuView]
          if (uu_view.nil?) || (uu_view[:sourceComponent].nil?)
            return nil
          end
          source_comp_code = uu_view[:sourceComponent]
          source_comp_code = source_comp_code.to_sym unless source_comp_code.nil?
          return container.component_by_code(source_comp_code)
        end


        # Serialize object to JSON in format corresponding to schema ues:SYSTEM:UU.OS.VUC/RESPONSE_MESSAGE-SCHEMA-V3
        #
        # @param [Hash] opt Optional parameters.
        # @see #JSON
        def to_json(opt = {})
          response = {}

          # add components
          comp_list = container.get_components
          comp_list.each do |k, v|
            response[k] = v.attributes
          end

          # add raw attributes
          @raw_attrs.each do |key, param|
            response[key] = param
          end

          # add messages
          msgs = container.messages
          if (!msgs.nil? && !msgs.empty?)
            view = create_view(response)
            view[:messages] = []
            msgs.each do |message|
              view[:messages] << message.raw_attrs
            end
          end

          # add focus if any
          focus = container.get_focus
          if focus
            view = create_view(response)
            view[:focus] = focus
          end

          # add view template
          unless @template_uri.nil?
            view = create_view(response)
            view[:templateUri] = @template_uri
          end
          response[:schemaUri] = RESPONSE_JSON_SCHEMA_URI
          return response.to_json(opt)
        end

        private

        # @private
        def create_view(response)
          view = response[:uuView]
          response[:uuView] = view = {} unless view
          return view
        end

        # @private
        def container
          @container ||= UU::OS::VUC::ComponentContainer.new(@ctx)
        end

      end
    end
  end
end
